from sqlparse import tokens, parse as sqlparse
from sqlparse.sql import Function, Case

from .sql_tokens import AliasableToken, SQLToken, SQLStatement, SQLCase
from ..exceptions import SQLDecodeError


class SQLFunc(AliasableToken):
    def __init__(self, *args):
        super().__init__(*args)

    @property
    def func_tok(self):
        if isinstance(self._token, Function):
            return self._token
        elif isinstance(self._token[0], Function):
            return self._token[0]
        else:
            raise SQLDecodeError('Not a function')

    @property
    def func(self) -> str:
        return self.func_tok.get_name()

    @property
    def parameters(self):
        return self.func_tok.get_parameters()

    @property
    def identifier(self):
        if not self.parameters[0]:
            return None
        return SQLToken.token2sql(self.parameters[0], self.query)

    @property
    def field(self):
        return self.alias if self.alias else self.query.token_alias.token2alias[self]

    @property
    def table(self):
        return self.identifier.table

    @property
    def column(self):
        return self.identifier.column

    @property
    def resolved_field(self):
        if self.case:
            return self.case.to_mongo()
        else:
            """FIX: FUNC('__col1')...FROM(SUBQUERY) syntax (field becomes '__col1.__col1')"""
            iden = self.identifier
            return f'${iden.column}' if iden.column == iden.table else f'${iden.field}'

    @property
    def is_distinct(self) -> str:
        """ to_mongo not handled yet """
        return self.func_tok[1][1].match(tokens.Keyword, 'DISTINCT')

    @property
    def case(self):
        t = self.func_tok[1][1]
        if not isinstance(t, Case):
            return None
        return SQLCase(t, self.query)

    @property
    def is_wildcard(self):
        return self.func_tok[1][1].ttype == tokens.Wildcard

    def to_mongo(self) -> dict:
        if self.func in ('MIN', 'MAX', 'SUM', 'AVG'):
            return {f'${self.func.lower()}': self.resolved_field}
        elif self.func == 'COUNT':
            if self.is_wildcard:
                return {'$sum': 1}
            return {
                '$sum': {
                    '$cond': {
                        'if': {
                            '$gt': [self.resolved_field, None]
                        },
                        'then': 1,
                        'else': 0
                    }
                }
            }
        else:
            raise SQLDecodeError(f'Unsupported func: {self.func}')



from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from logging import getLogger

logger = getLogger(__name__)


class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):

    sql_create_index = "ALTER TABLE %(table)s ADD CONSTRAINT %(name)s INDEX (%(columns)s)%(extra)s"
    sql_delete_index = "ALTER TABLE %(table)s DROP CONSTRAINT %(name)s INDEX"
    sql_create_unique_index = "ALTER TABLE %(table)s ADD CONSTRAINT %(name)s UNIQUE (%(columns)s)(%(condition)s)"

    def quote_value(self, value):
        return value

    def prepare_default(self, value):
        raise NotImplementedError()




import abc
import re
from functools import cached_property
from typing import Union as U, Iterator, Optional as O, Optional, List

from pymongo import ASCENDING, DESCENDING
from sqlparse import tokens, parse as sqlparse
from sqlparse.sql import Token, Identifier, Function, Comparison, Parenthesis, IdentifierList, Statement, Operation, \
    Case

from . import query as query_module
from ..exceptions import SQLDecodeError, NotSupportedError

all_token_types = U['SQLConstIdentifier',
                    'djongo.sql2mongo.functions.CountFunc',
                    'djongo.sql2mongo.functions.SimpleFunc',
                    'SQLIdentifier',
                    'SQLComparison',
                    'SQLPlaceholder',
                    'SQLOperation']


def strip_alias(sql_token, tok_value):
    if isinstance(sql_token, SQLOperation) and sql_token.has_explicit_alias or not isinstance(
            sql_token, SQLOperation):
        unaliased = tok_value.rsplit(' ', 1)[0]
        unaliased = unaliased.rsplit(' ', 1)[0] if unaliased.upper().endswith(' AS') else unaliased
    else:
        unaliased = sql_token._token.value
    return unaliased


class SQLToken:

    @abc.abstractmethod
    def __init__(self,
                 token: Token,
                 query: 'query_module.BaseQuery'):
        self._token = token
        self.query = query
        self.uses_existing_fields = False

    def __repr__(self):
        return f'{self._token}'

    @staticmethod
    def tokens2sql(token: Token,
                   query: 'query_module.BaseQuery'
                   ) -> Iterator[all_token_types]:
        from .functions import SQLFunc
        if isinstance(token, Identifier):
            # Bug fix for sql parse
            if isinstance(token[0], Parenthesis):
                if isinstance(token[0][1], Operation):
                    yield SQLOperation(token, query)
                else:
                    try:
                        int(token[0][1].value)
                    except ValueError:
                        yield SQLIdentifier(token[0][1], query)
                    else:
                        yield SQLConstIdentifier(token, query)
            elif isinstance(token[0], Function):
                yield SQLFunc(token, query)
            else:
                yield SQLIdentifier(token, query)
        elif isinstance(token, Function):
            yield SQLFunc(token, query)
        elif isinstance(token, Comparison):
            yield SQLComparison(token, query)
        elif isinstance(token, IdentifierList):
            for tok in token.get_identifiers():
                yield from SQLToken.tokens2sql(tok, query)
        elif isinstance(token, Parenthesis):
            if isinstance(token[1], Operation):
                yield SQLOperation(token[1], query)
            else:
                yield SQLPlaceholder(token, query)
        elif token.ttype == tokens.Name.Placeholder:
            yield SQLPlaceholder(token, query)
        elif isinstance(token, Operation):
            yield SQLOperation(token, query)
        elif isinstance(token, Case):
            yield SQLCase(token, query)
        else:
            raise SQLDecodeError(f'Unsupported: {token.value}')

    @staticmethod
    def token2sql(token: Token,
                  query: 'query_module.BaseQuery'
                  ) -> all_token_types:
        return next(SQLToken.tokens2sql(token, query))

    @staticmethod
    def placeholder_index(token) -> int:
        # only matches from start of token
        match = re.match(r'%\(([0-9]+)\)s', token.value, flags=re.IGNORECASE)
        return int(match.group(1)) if match else None


class AliasableToken(SQLToken):

    @abc.abstractmethod
    def __init__(self, *args):
        super().__init__(*args)
        self.token_alias: 'query_module.TokenAlias' = self.query.token_alias

        try:
            self._ord = ORDER_BY_MAP[self._token.get_ordering()]
            self._token = self._token[0]
        except Exception:
            self._ord = None

        tok_value = self._token.value
        self.unaliased = strip_alias(self, tok_value)
        self.is_array_len = [f for f in self.unaliased.split('.')][-1] == '"len"'

        if self.alias:
            self.token_alias.alias2token[self.alias] = self
            self.token_alias.token2alias[self] = self.alias
            if self.has_explicit_alias:
                self.token_alias.aliased_names.add(self.alias)

    def __hash__(self):
        if self.has_explicit_alias:
            return hash(self._token[0].value)
        return hash(self._token.value)

    def __eq__(self, other):
        return hash(self) == hash(other)

    @property
    def has_explicit_alias(self):
        return len(self._token.tokens) >= 5 and self._token[-3].match(tokens.Keyword, 'AS')

    @property
    def alias(self) -> Optional[str]:
        # bug fix sql parse
        try:
            return self._token.get_alias()
        except Exception:
            return None

    @property
    def order(self):
        if self._ord is None:
            raise SQLDecodeError
        return self._ord


class SQLIdentifier(AliasableToken):

    def __init__(self, *args):
        super().__init__(*args)

    @cached_property
    def field(self) -> str:
        if self.given_table in self.query.token_alias.aliased_names:
            return self.given_table

        if self.table == self.query.left_table:
            return self.column
        else:
            return f'{self.table}.{self.column}'

    @cached_property
    def table(self) -> str:
        name = self.given_table
        alias2token = self.token_alias.alias2token
        try:
            return alias2token[name].table
        except KeyError:
            return name

    @cached_property
    def given_table(self) -> str:
        name = self._token.get_parent_name()
        if name is None:
            name = self._parse_token_value[0]

        if name is None:
            raise SQLDecodeError
        return name

    @cached_property
    def column(self) -> str:
        name = self._parse_token_value[1]
        if name is None:
            raise SQLDecodeError
        return name

    def to_mongo(self):
        from .operators import parse_field
        return parse_field(self.column, self.is_array_len)

    @cached_property
    def _parse_token_value(self):
        # fix for JSONField accessors (example: a.b.c.d)
        tokens = [f for f in self.unaliased.split('.')]
        if tokens[-1] == '"len"':
            tokens.pop()
        table, column = tokens[0], '.'.join(tokens[1:])
        if not column:
            column = table
        table, column = table.replace('"', ''), column.replace('"', '')
        return table, column


class SQLOperation(AliasableToken):
    def __init__(self, *args):
        super().__init__(*args)
        statement = sqlparse(self.unaliased)[0]
        self.token: Token = statement[0]
        if isinstance(self.token, Parenthesis):  # unwrap parenthesis
            self.token = self.token[1]
        self._left = SQLToken.token2sql(self.token[0], self.query)
        self._right = SQLToken.token2sql(self.token[-1], self.query)
        self._table = None

        if isinstance(self._left, SQLIdentifier) or isinstance(self._right, SQLIdentifier):
            self.uses_existing_fields = True
            ltable = getattr(self._left, 'table', '')
            rtable = getattr(self._right, 'table', '')
            if ltable and rtable and ltable != rtable:
                raise SQLDecodeError('Unable to do SQL operation on fields from different tables yet.')
            self._table = ltable or rtable
        else:
            self.uses_existing_fields = getattr(
                self._left, 'uses_existing_fields', False) or getattr(self._right, 'uses_existing_fields', False)
            self._table = getattr(self._left, '_table', False) or getattr(self._right, '_table', False)

    @property
    def left(self):
        return self._left

    @property
    def right(self):
        return self._right

    @property
    def operator(self):
        from .operators import OPERATOR_MAP
        if self.alias:
            value = self._token[0][1][2].value
        else:
            value = self._token[2].value
        return OPERATOR_MAP[value]

    @property
    def field(self):
        return self.to_mongo()

    @property
    def table(self) -> str:
        return self._table

    def to_mongo(self):
        return {self.operator: [self.left.to_mongo(), self.right.to_mongo()]}


class SQLConstIdentifier(AliasableToken):

    def __init__(self, *args):
        super().__init__(*args)

    @property
    def value(self) -> int:
        return int(self._token[0][1].value)

    def to_mongo(self) -> dict:
        return {'$literal': self.value}


class SQLComparison(SQLToken):
    def __init__(self, *args):
        super().__init__(*args)
        l = self._token.left
        r = self._token.right

        # Evaluate rhs of a comparison token, used in update statement
        # - have to eagerly evaluate to get uses_existing_fields
        self._rhs = None
        if isinstance(r, Identifier):
            self._rhs = f"${self.right_column}"
        elif isinstance(r, Case):
            op = SQLCase(r, self.query)
            self.uses_existing_fields = op.uses_existing_fields
            self._rhs = op.to_mongo()
        elif isinstance(r, Parenthesis) and isinstance(r[1], Operation):
            op = SQLOperation(r[1], self.query)
            self.uses_existing_fields = op.uses_existing_fields
            self._rhs = op.to_mongo()
        elif isinstance(r, Operation):
            op = SQLOperation(r, self.query)
            self.uses_existing_fields = op.uses_existing_fields
            self._rhs = op.to_mongo()
        elif r.ttype == tokens.Name.Placeholder:
            self._rhs = self.query.params[self.placeholder_index(r)]
        elif r.match(tokens.Keyword, 'NULL'):
            self._rhs = None

    @property
    def left_table(self):
        lhs = SQLIdentifier(self._token.left, self.query)
        return lhs.table

    @property
    def left_column(self):
        lhs = SQLIdentifier(self._token.left, self.query)
        return lhs.column

    @property
    def right_table(self):
        rhs = SQLIdentifier(self._token.right, self.query)
        return rhs.table

    @property
    def right_column(self):
        rhs = SQLIdentifier(self._token.right, self.query)
        return rhs.column

    @property
    def rhs(self):
        return self._rhs


class SQLPlaceholder(SQLToken):
    def __iter__(self):
        tok = self._token[1:-1][0]
        if isinstance(tok, IdentifierList):
            for aid in tok.get_identifiers():
                yield self.get_value(aid)

        else:
            yield self.get_value(tok)

    def __init__(self, token: Token, query: 'query_module.BaseQuery'):
        super().__init__(token, query)

    def get_value(self, tok: Token = None):
        tok = tok if tok else self._token
        if tok.ttype == tokens.Name.Placeholder:
            return self.placeholder_index(tok)
        elif tok.match(tokens.Keyword, 'NULL'):
            return None
        elif tok.match(tokens.Keyword, 'DEFAULT'):
            return 'DEFAULT'
        else:
            raise SQLDecodeError

    def to_mongo(self):
        return self.query.params[self.get_value()]


class SQLCase(SQLToken):
    def __init__(self, *args):
        if not isinstance(args[0], Case):
            raise SQLDecodeError
        super().__init__(*args)
        # assume true, could fail in edge cases, e.g. case when NOW() > %(1)s then %(2)s ...
        self.uses_existing_fields = True

    @property
    def cases(self):
        return self._token.get_cases(skip_ws=True)

    def to_mongo(self):
        from .operators import WhereOp
        cases = []
        for c in self.cases:
            case_toks, then_toks = c
            case_val, then_val = None, None
            if case_toks:
                case_stmt = SQLStatement(sqlparse(' '.join(str(s) for s in case_toks))[0])
                case_op = WhereOp(statement=case_stmt, query=self.query, params=self.query.params)
                case_val = case_op.to_mongo()
            if then_toks:
                then_stmt = SQLStatement(sqlparse(' '.join(str(s) for s in then_toks[1:]))[0])
                then_op = SQLToken.token2sql(then_stmt.next(), query=self.query)
                then_val = then_op.to_mongo()
            cases.append((case_val, then_val))
        if not cases:
            raise SQLDecodeError
        cases_to_mongo = {'$cond': [cases[0][0], cases[0][1]]}
        cond = cases_to_mongo['$cond']
        for c in cases[1:]:
            if c[0]:  # WHEN..THEN
                nested_cond = {'$cond': [c[0], c[1]]}
                cond.append(nested_cond)
                cond = nested_cond['$cond']
            else:  # ELSE
                cond.append(c[1])
        return cases_to_mongo


class SQLStatement:

    @property
    def current_token(self) -> Token:
        return self._statement[self._tok_id]

    def __init__(self, statement: U[Statement, Token]):
        self._statement = statement
        self._tok_id = 0
        self._gen_inst = self._generator()

    def __getattr__(self, item):
        return getattr(self._statement, item)

    def __iter__(self) -> Token:
        yield from self._gen_inst

    def __repr__(self):
        return str(self._statement)

    def __getitem__(self, item: slice):
        start = (item.start or 0) + self._tok_id
        stop = item.stop and self._tok_id + item.stop
        sql = ''.join(str(tok) for tok in self._statement[start:stop])
        sql = sqlparse(sql)[0]
        return SQLStatement(sql)

    def next(self) -> O[Token]:
        # self._tok_id, token = self._statement.token_next(self._tok_id)
        try:
            return next(self._gen_inst)
        except StopIteration:
            return None

    def skip(self, num):
        self._tok_id += num

    @property
    def prev_token(self) -> Token:
        return self._statement.token_prev(self._tok_id)[1]

    @property
    def next_token(self) -> Token:
        return self._statement.token_next(self._tok_id)[1]

    def _generator(self):
        token = self._statement[self._tok_id]
        while self._tok_id is not None:
            yield token
            self._tok_id, token = self._statement.token_next(self._tok_id)


class SQLColumnDef:
    not_null = object()
    unique = object()
    autoincrement = object()
    primarykey = object()
    null = object()
    _map = {
        'UNIQUE': unique,
        'AUTOINCREMENT': autoincrement,
        'PRIMARY KEY': primarykey,
        'NOT NULL': not_null,
        'NULL': null
    }
    supported_data_types = None

    def __init__(self,
                 name: str = None,
                 data_type: str = None,
                 col_constraints: set = None):
        self.name = name
        self.data_type = data_type
        self.col_constraints = col_constraints

    @staticmethod
    def _get_constraints(others: str):
        while others:
            try:
                name, others = others.split(' ', 1)
            except ValueError:
                name = others
                others = None
            try:
                yield SQLColumnDef._map[name]
            except KeyError:
                if others:
                    try:
                        part2, others = others.split(' ', 1)
                    except ValueError:
                        part2 = others
                        others = None

                    name = f'{name} {part2}'
                    try:
                        yield SQLColumnDef._map[name]
                    except KeyError:
                        raise SQLDecodeError(f'Unknown column constraint: {name}')
                else:
                    raise SQLDecodeError(f'Unknown column constraint: {name}')

    @staticmethod
    def sql2col_defs(sql: str):
        sql = sql[1:-1]
        while sql:
            if sql[0] == '"':
                try:
                    def_str, sql = sql.split(',', 1)
                except ValueError:
                    def_str = sql
                    sql = None
                else:
                    sql = sql.strip()
                yield SQLColumnDef.def_str2col_def(def_str)
            elif sql.startswith('CONSTRAINT'):
                # Temporary hack
                indexes = [None]
                i = 0
                for i, c in enumerate(sql):
                    if c == '(':
                        if indexes[0] is None:
                            indexes.pop()
                        indexes.append(i)
                    elif c == ')':
                        indexes.pop()
                    if not indexes:
                        break
                if len(sql[i:]) > 1:
                    sql = sql[i + 3:]
                else:
                    sql = None
                yield SQLColumnConstraint()
            else:
                raise SQLDecodeError(f'Syntax Error: {sql}')

    @classmethod
    def def_str2col_def(cls, def_str: str):
        if not cls.supported_data_types:
            from djongo.base import DatabaseWrapper
            cls.supported_data_types = set(DatabaseWrapper.data_types.values())

        name, other = def_str[1:].split('"', 1)
        other = other.strip()

        data_type, constraint_sql = other.split(' ', 1)
        if data_type not in cls.supported_data_types:
            raise NotSupportedError(f'Data of type: {data_type}')

        col_constraints = set(SQLColumnDef._get_constraints(constraint_sql))
        return SQLColumnDef(name=name,
                            data_type=data_type,
                            col_constraints=col_constraints)

    @classmethod
    def statement2col_defs(cls, token: Token):
        from djongo.base import DatabaseWrapper
        supported_data_types = set(DatabaseWrapper.data_types.values())

        defs = token.value.strip('()').split(',')
        for col in defs:
            col = col.strip()
            name, other = col.split(' ', 1)
            if name == 'CONSTRAINT':
                yield SQLColumnConstraint()
            else:
                if col[0] != '"':
                    raise SQLDecodeError('Column identifier not quoted')
                name, other = col[1:].split('"', 1)
                other = other.strip()

                data_type, constraint_sql = other.split(' ', 1)
                if data_type not in supported_data_types:
                    raise NotSupportedError(f'Data of type: {data_type}')

                col_constraints = set(SQLColumnDef._get_constraints(constraint_sql))
                yield SQLColumnDef(name=name,
                                   data_type=data_type,
                                   col_constraints=col_constraints)


class SQLColumnConstraint(SQLColumnDef):
    pass


ORDER_BY_MAP = {
    'ASC': ASCENDING,
    'DESC': DESCENDING
}



from logging import getLogger
from pymongo import MongoClient

logger = getLogger(__name__)
clients = {}


def connect(db, **kwargs):
    try:
        return clients[db]
    except KeyError:
        logger.debug('New MongoClient connection')
        clients[db] = MongoClient(**kwargs, connect=False)
    return clients[db]


class Error(Exception):  # NOQA: StandardError undefined on PY3
    pass


class InterfaceError(Error):
    pass


class DatabaseError(Error):
    pass


class DataError(DatabaseError):
    pass


class OperationalError(DatabaseError):
    pass


class IntegrityError(DatabaseError):
    pass


class InternalError(DatabaseError):
    pass


class ProgrammingError(DatabaseError):
    pass


class NotSupportedError(DatabaseError):
    pass


def Binary(value):
    return value



import collections
import datetime

import bson
from django.db.backends.base.introspection import BaseDatabaseIntrospection, FieldInfo, TableInfo
from django.db.models import Index


class DatabaseIntrospection(BaseDatabaseIntrospection):
    SAMPLE_SIZE = 10000
    TYPE_MAPPING = {
        int: bson.int64.Int64,
    }

    data_types_reverse = {
        bson.int64.Int64: 'BigIntegerField',
        bson.objectid.ObjectId: 'ObjectIdField',
        collections.OrderedDict: 'JSONField',
        datetime.date: 'DateField',
        datetime.datetime: 'DateTimeField',
        bool: 'BooleanField',
        dict: 'JSONField',
        float: 'FloatField',
        int: 'IntegerField',
        list: 'JSONField',
        str: 'CharField',
        'text': 'TextField',
        'int64': 'BigIntegerField',
        'int32': 'IntegerField',
        'number': 'DecimalField',
        'string': 'CharField',
        'boolean': 'BooleanField',
        'object': 'djongo.models.DictField',
        'array': 'djongo.models.ListField',
        'oid': 'djongo.models.ObjectIdField',
        'date': 'DateTimeField'

    }

    # def table_names(self, cursor=None, include_views=False):
    #     return sorted(cursor.m_cli_connection.collection_names(False))

    def get_table_list(self, cursor):

        return [
            TableInfo(c, 't')
            for c in cursor.db_conn.list_collection_names()
            if c != '__schema__'
        ]

    def get_constraints(self, cursor, table_name):
        constraint = {}

        indexes = cursor.db_conn[table_name].index_information()
        for name, info in indexes.items():
            if name == '_id_':
                continue

            columns = [field[0] for field in info['key']]
            orders = ['ASC' if field[1] == 1 else 'DESC'
                      for field in info['key']]
            constraint[name] = {
                'columns': columns,
                'primary_key': name == '__primary_key__',
                'unique': info.get('unique', False),
                'index': True,
                'orders': orders,
                "foreign_key": False,
                "check": False,
                'type': Index.suffix
            }
        return constraint

    def get_key_columns(self, cursor, table_name):
        return []

    def get_indexes(self, cursor, table_name):
        return self.get_constraints(cursor, table_name)

    def get_relations(self, cursor, table_name):
        return []

    def get_sequences(self, cursor, table_name, table_fields=()):
        pk_col = self.get_primary_key_column(cursor, table_name)
        return [{'table': table_name, 'column': pk_col}]

    def get_table_description(self, cursor, table_name):
        colspecs = collections.defaultdict(lambda: dict(
            types=collections.Counter(),
            specs=collections.defaultdict(int),
        ))
        fields = cursor.db_conn['__schema__'].find_one(
            {'name': table_name},
            {'fields': True}
        )['fields']
        columns = []
        for name, properties in fields.items():
            columns.append(
                FieldInfo(
                    name=name,
                    type_code=properties['type_code'],
                    display_size=None,
                    internal_size=None,
                    precision=None,
                    scale=None,
                    null_ok=None,
                    default=None
                )
            )
        return columns


import re
import functools


encode_dict = {
    ".": "\\u002e",
    "\$": "\\u0024",
    "\\": "\\\\",
}
encode_regex = re.compile("|".join(map(re.escape, encode_dict.keys())))


@functools.lru_cache(maxsize=1024)
def encode_key(s: str):
    if isinstance(s, int):
        return str(s)
    elif isinstance(s, str):
        return encode_regex.sub(lambda match: encode_dict[match.group()], s)
    else:
        return s


def encode_keys(obj, is_key=False):
    if is_key:
        if isinstance(obj, (int, str)):
            return encode_key(obj)
        return obj
    if isinstance(obj, dict):
        return {encode_keys(k, True): encode_keys(v) for k, v in obj.items()}
    if isinstance(obj, list):
        return [encode_keys(v) for v in obj]
    return obj


decode_dict = {
    "\\u002e": ".",
    "\\u0024": "\$",
    "\\\\": "\\",
}
decode_regex = re.compile("|".join(map(re.escape, decode_dict.keys())))


@functools.lru_cache(maxsize=1024)
def decode_key(s: str):
    if s.isdigit() or (s.startswith('-') and s[1:].isdigit()):
        return int(s)
    else:
        return decode_regex.sub(lambda match: decode_dict[match.group()], s)


def decode_keys(obj, is_key=False):
    if isinstance(obj, str):
        if is_key:
            return decode_key(obj)
        return obj
    if isinstance(obj, dict):
        return {decode_keys(k, True): decode_keys(v) for k, v in obj.items()}
    if isinstance(obj, list):
        return [decode_keys(v) for v in obj]
    return obj



import re
import typing
from collections import defaultdict
from itertools import chain

from sqlparse import tokens
from sqlparse.sql import Token, Parenthesis, Comparison, IdentifierList, Identifier, Function

from djongo.utils import encode_key
from . import query
from .functions import SQLFunc
from .sql_tokens import SQLToken, SQLStatement, SQLComparison, SQLIdentifier
from ..exceptions import SQLDecodeError


def re_index(value: str):
    match = re.match(r'%\(([0-9]+)\)s', value, flags=re.IGNORECASE)
    if match:
        index = int(match.group(1))
    else:
        match = re.match(r'NULL', value, flags=re.IGNORECASE)
        if not match:
            raise SQLDecodeError('Unable to find placeholder or NULL')
        index = None
    return index


def set_constant(op, token, attr='_constant', unwrap_dict=True):
    if not isinstance(token, Token):
        raise SQLDecodeError(f'{token} is not a Token!')
    index = re_index(token.value)
    setattr(op, attr, op.params[index] if index is not None else None)
    const_val = getattr(op, attr, None)
    if unwrap_dict and isinstance(const_val, dict):  # unwrap dict
        key, value = next(iter(const_val.items()))
        setattr(op, '_field_ext', key)
        setattr(op, attr, value)
    else:
        setattr(op, '_field_ext', None)


def parse_field(val, is_array_len=False, with_dollar=True):
    val = f'${val}' if with_dollar and isinstance(val, str) else val
    return {'$size': {"$ifNull": [val, []]}} if is_array_len else val


class _Op:

    def __init__(
            self,
            statement: SQLStatement,
            query: 'query.SelectQuery',
            params: tuple = None,
            name='generic'):
        super().__init__()
        self.lhs: typing.Optional[_Op] = None
        self.rhs: typing.Optional[_Op] = None

        if params is not None:
            self.params = params
        else:
            self.params = query.params
        self.query = query
        self.left_table = query.left_table

        self.statement = statement
        self.is_negated = False
        self._name = name
        self.precedence = OPERATOR_PRECEDENCE[name]

    def negate(self):
        raise NotImplementedError

    def to_mongo(self):
        raise NotImplementedError


class _UnaryOp(_Op):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._op = None

    def negate(self):
        raise NotImplementedError

    def evaluate(self):
        self.rhs.evaluate()

    def to_mongo(self):
        return self.rhs.to_mongo()


class _BinaryOp(_Op):

    def __init__(self, *args, token='prev_token', **kwargs):
        super().__init__(*args, **kwargs)
        tok = SQLToken.token2sql(getattr(self.statement, token), self.query)
        if isinstance(tok, (SQLIdentifier, SQLFunc)):
            self._field = tok.field
        elif isinstance(tok, SQLComparison):
            self._field = SQLIdentifier(tok._token.left, tok.query).field
        else:
            raise SQLDecodeError(f'Unexpected token: {tok}')

    def negate(self):
        raise SQLDecodeError('Negating IN/NOT IN not supported')

    def to_mongo(self):
        raise NotImplementedError

    def evaluate(self):
        pass


class _InNotInOp(_BinaryOp):
    @property
    def is_in(self):
        return (isinstance(self, InOp) and not self.is_negated) or (isinstance(self, NotInOp) and self.is_negated)

    def _fill_in(self, token):
        self._in = []

        # Check for nested
        if token[1].ttype == tokens.DML:
            from .converters import NestedInQueryConverter

            self.query.nested_query = NestedInQueryConverter(token, self.query, 0)
            return

        for index in SQLToken.token2sql(token, self.query):
            if index is not None:
                self._in.append(self.params[index])
            else:
                self._in.append(None)

    def negate(self):
        self.is_negated = True

    def to_mongo(self):
        if self.query.nested_query is not None and not self._in:
            expr = {'$in': [f'${self._field}', '$_nested_in']}
            expr = {'$not': expr} if not self.is_in else expr
            return {'$expr': expr}
        else:
            expr = {'$in': self._in}
            expr = {'$not': expr} if not self.is_in else expr
            return {self._field: expr}


class NotInOp(_InNotInOp):
    def __init__(self, *args, token='prev_token', **kwargs):
        super().__init__(name='NOT IN', *args, token=token, **kwargs)
        self._fill_in(getattr(self.statement, token)[-1])


class InOp(_InNotInOp):
    def __init__(self, *args, token='prev_token', **kwargs):
        super().__init__(name='IN', *args, token=token, **kwargs)
        self._fill_in(getattr(self.statement, token)[-1])


class LikeOp(_BinaryOp):

    def __init__(self, *args, token='current_token', **kwargs):
        super().__init__(name='LIKE', *args, token=token, **kwargs)
        self._regex = None
        self._make_regex(getattr(self.statement, token))

    def _make_regex(self, token):
        index = SQLToken.placeholder_index(token.right)

        to_match = self.params[index]
        if isinstance(to_match, dict):
            field_ext, to_match = next(iter(to_match.items()))
            self._field += '.' + field_ext
        if not isinstance(to_match, str):
            raise SQLDecodeError
        # Ensure all regex special characters are handled
        to_match = re.escape(to_match)
        # Like expression special character - the expression above will escape \\
        to_match = to_match.replace('\\\\_', '_')
        # Like expression special character + ensure normal % is handled properly
        to_match = '%'.join(s.replace('%', '.*') for s in to_match.split('\\\\%'))
        to_match = f"^{to_match}$"
        self._regex = to_match

    def to_mongo(self):
        return {self._field: {'$regex': self._regex}}


class iLikeOp(LikeOp):
    def to_mongo(self):
        return {self._field: {
            '$regex': self._regex,
            '$options': 'im'
        }}


class RegexpOp(_BinaryOp):

    def __init__(self, *args, token='current_token', **kwargs):
        super().__init__(name='REGEXP', *args, token=token, **kwargs)
        self._regex = None
        self._make_regex(self.statement.next())

    def _make_regex(self, token):
        index = SQLToken.placeholder_index(token)

        to_match = self.params[index]
        if isinstance(to_match, dict):
            field_ext, to_match = next(iter(to_match.items()))
            self._field += '.' + field_ext
        if not isinstance(to_match, str):
            raise SQLDecodeError
        self._regex = to_match

    def to_mongo(self):
        return {self._field: {'$regex': self._regex}}


class iRegexpOp(RegexpOp):
    def to_mongo(self):
        return {self._field: {'$regex': self._regex, '$options': 'im'}}


class IsOp(_BinaryOp):

    def __init__(self, *args, **kwargs):
        super().__init__(name='IS', *args, **kwargs)
        token = self.statement
        key = token.next()
        if key.match(tokens.Keyword, 'Null'):
            self._is_null = True
        elif key.match(tokens.Keyword, 'Not null'):
            self._is_null = False
        else:
            raise SQLDecodeError

    def negate(self):
        self.is_negated = True

    def to_mongo(self):
        is_null = not self._is_null if self.is_negated else self._is_null
        return {
            self._field: None if is_null else {'$ne': None}
        }


class BetweenOp(_BinaryOp):

    def __init__(self, *args, **kwargs):
        super().__init__(name='BETWEEN', *args, **kwargs)
        token = self.statement

        lower = token.next()
        lower = SQLToken.placeholder_index(lower)
        self._lower = self.params[lower]

        _and = token.next()
        if not _and.match(tokens.Keyword, 'AND'):
            raise SQLDecodeError

        upper = token.next()
        upper = SQLToken.placeholder_index(upper)
        self._upper = self.params[upper]

    def negate(self):
        self.is_negated = True

    def to_mongo(self):
        if not self.is_negated:
            return {
                self._field: {
                    '$gte': self._lower,
                    '$lte': self._upper
                }
            }
        else:
            return {
                self._field: {
                    '$not': {
                        '$gte': self._lower,
                        '$lte': self._upper
                    }
                }
            }


class NotOp(_UnaryOp):
    def __init__(self, *args, **kwargs):
        super().__init__(name='NOT', *args, **kwargs)

    def negate(self):
        raise SQLDecodeError

    def evaluate(self):
        self.rhs.negate()
        if isinstance(self.rhs, ParenthesisOp):
            self.rhs.evaluate()
        if self.lhs is not None:
            self.lhs.rhs = self.rhs


class _AndOrOp(_Op):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._acc = []

    def negate(self):
        self.is_negated = True

    def op_type(self):
        raise NotImplementedError

    def evaluate(self):
        if not (self.lhs and self.rhs):
            raise SQLDecodeError

        if isinstance(self.lhs, _AndOrOp):
            if self.op_type() == self.lhs.op_type():
                self._acc = self.lhs._acc + self._acc
            else:
                self._acc.insert(0, self.lhs)

        elif isinstance(self.lhs, ParenthesisOp):
            self.lhs.evaluate()
            self._acc.append(self.lhs)

        elif isinstance(self.lhs, _Op):
            self._acc.append(self.lhs)

        else:
            raise SQLDecodeError

        if isinstance(self.rhs, _AndOrOp):
            if self.op_type() == self.rhs.op_type():
                self._acc.extend(self.rhs._acc)
            else:
                self._acc.append(self.rhs)

        elif isinstance(self.rhs, ParenthesisOp):
            self.rhs.evaluate()
            self._acc.append(self.rhs)

        elif isinstance(self.rhs, _Op):
            self._acc.append(self.rhs)

        else:
            raise SQLDecodeError

        if self.lhs.lhs is not None:
            self.lhs.lhs.rhs = self
        if self.rhs.rhs is not None:
            self.rhs.rhs.lhs = self

    def to_mongo(self):
        if self.op_type() == AndOp:
            oper = '$and'
        else:
            oper = '$or'

        docs = [itm.to_mongo() for itm in self._acc]
        return {oper: docs}


class AndOp(_AndOrOp):

    def __init__(self, *args, **kwargs):
        super().__init__(name='AND', *args, **kwargs)

    def op_type(self):
        if not self.is_negated:
            return AndOp
        else:
            return OrOp


class OrOp(_AndOrOp):

    def __init__(self, *args, **kwargs):
        super().__init__(name='OR', *args, **kwargs)

    def op_type(self):
        if not self.is_negated:
            return OrOp
        else:
            return AndOp


class _StatementParser:

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._ops: typing.List[_Op] = []
        self._cmp_ops: typing.List[_Op] = []
        self._op = None

    def _token2op(self,
                  tok: Token,
                  statement: SQLStatement) -> '_Op':
        op = None
        kw = {'statement': statement, 'query': self.query}
        if tok.match(tokens.Name.Placeholder, '%(.+)s', regex=True):
            return op

        if tok.match(tokens.Keyword, 'AND'):
            op = AndOp(**kw)

        elif tok.match(tokens.Keyword, 'OR'):
            op = OrOp(**kw)

        elif any(t.match(tokens.Comparison, 'NOT IN') for t in [tok, *getattr(tok, 'tokens', [])]):
            op = NotInOp(**kw, token='current_token')

        elif any(t.match(tokens.Comparison, 'IN') for t in [tok, *getattr(tok, 'tokens', [])]):
            op = InOp(**kw, token='current_token')

        elif tok.match(tokens.Keyword, 'NOT'):
            if statement.next_token.match(tokens.Keyword, 'IN'):
                op = NotInOp(**kw)
                statement.skip(1)
            else:
                op = NotOp(**kw)
        elif tok.value.endswith('eREGEXP'):
            op = RegexpOp(**kw)
        elif tok.value.endswith('iREGEXP'):
            op = iRegexpOp(**kw)

        elif isinstance(tok, Comparison) and 'iLIKE' in tok.normalized:
            op = iLikeOp(**kw)

        elif isinstance(tok, Comparison) and 'LIKE' in tok.normalized:
            op = LikeOp(**kw)

        elif tok.match(tokens.Keyword, 'BETWEEN'):
            op = BetweenOp(**kw)
            statement.skip(3)

        elif tok.match(tokens.Keyword, 'IS'):
            op = IsOp(**kw)

        elif isinstance(tok, Function) and statement.next_token.match(tokens.Keyword, 'IS'):
            statement.skip(2)
            op = IsOp(**kw)

        elif tok.value in JSON_OPERATORS:
            op = JSONOp(**kw)

        elif isinstance(tok, Comparison):
            op = CmpOp(tok, self.query)

        elif isinstance(tok, Parenthesis):
            if (tok[1].match(tokens.Name.Placeholder, '.*', regex=True)
                    or tok[1].match(tokens.Keyword, 'Null')
                    or isinstance(tok[1], IdentifierList)
                    or tok[1].ttype == tokens.DML
            ):
                pass
            else:
                op = ParenthesisOp(SQLStatement(tok), self.query)

        elif tok.match(tokens.Punctuation, (')', '(')):
            pass

        elif isinstance(tok, Identifier):
            t = statement.next_token
            if not t or t.match(tokens.Punctuation, (')', '(')) or t.match(tokens.Keyword, ('AND', 'OR')):
                op = ColOp(tok, self.query)
        else:
            raise SQLDecodeError

        return op

    def _statement2ops(self):
        def link_op():
            if prev_op is not None:
                prev_op.rhs = op
                op.lhs = prev_op

        statement = self.statement

        prev_op = None
        op = None
        for tok in statement:
            op = self._token2op(tok, statement)
            if not op:
                continue
            link_op()
            if isinstance(op, CmpOp):
                self._cmp_ops.append(op)
            if not isinstance(op, (CmpOp, ParenthesisOp, ColOp)):
                self._op_precedence(op)
            prev_op = op

        if prev_op.lhs is None:
            if isinstance(prev_op, (CmpOp, ParenthesisOp, ColOp)):
                self._ops.append(prev_op)

    def _op_precedence(self, operator: _Op):
        ops = self._ops
        if not ops:
            ops.append(operator)
            return

        for i in range(len(ops)):
            if operator.precedence > ops[i].precedence:
                ops.insert(i, operator)
                break
        else:
            ops.append(operator)

    def evaluate(self):
        if self._op is not None:
            return

        if not self._ops:
            raise SQLDecodeError

        op = None
        for op in self._ops:
            op.evaluate()
        self._ops.clear()
        self._op = op


class WhereOp(_Op, _StatementParser):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.statement.skip(2)
        self._statement2ops()
        self.evaluate()

    def negate(self):
        raise NotImplementedError

    def to_mongo(self):
        return self._op.to_mongo()


class ParenthesisOp(_Op, _StatementParser):

    def to_mongo(self):
        return self._op.to_mongo()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._statement2ops()

    def negate(self):
        for op in chain(self._ops, self._cmp_ops):
            op.negate()


class CmpOp(_Op):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._identifier = SQLToken.token2sql(self.statement.left, self.query)
        self._operator = OPERATOR_MAP[self.statement.token_next(0)[1].value]

        if isinstance(self.statement.right, Identifier):
            # Comparing against field in same doc vs a constant.
            self._is_right_constant = False
            self._right_identifier = SQLToken.token2sql(self.statement.right, self.query)
            if self._right_identifier.table != self._identifier.table:
                raise SQLDecodeError('Join using WHERE not supported')
        else:
            # Comparing with a constant
            self._is_right_constant = True
            set_constant(self, self.statement.right)

    def negate(self):
        self.is_negated = True

    def evaluate(self):
        pass

    def _compare_fields_to_mongo(self):
        """Comparison for fields in same document.
        https://docs.mongodb.com/manual/reference/operator/query/expr/#behavior
        """
        l_field = self.query.token_alias.similartoken2alias(self._identifier)
        l_field = f'${l_field}' if l_field else parse_field(self._identifier.field, self._identifier.is_array_len)
        r_field = self.query.token_alias.similartoken2alias(self._right_identifier)
        r_field = f'${r_field}' if r_field else parse_field(
            self._right_identifier.field, self._right_identifier.is_array_len)
        if not self.is_negated:
            return {'$expr': {self._operator: [l_field, r_field]}}
        else:
            return {'$expr': {'$not': {self._operator: [l_field, r_field]}}}

    def _compare_constant_to_mongo(self):
        field = self.query.token_alias.similartoken2alias(self._identifier)
        field = f'${field}' if field else parse_field(self._identifier.field, self._identifier.is_array_len)
        constant = parse_field(self._constant, False, with_dollar=False)
        if self._field_ext:
            field += '.' + self._field_ext
        if not self.is_negated:
            return {'$expr': {self._operator: [field, constant]}}
        else:
            return {'$expr': {'$not': {self._operator: [field, constant]}}}

    def to_mongo(self):
        if self._is_right_constant:
            return self._compare_constant_to_mongo()
        else:
            return self._compare_fields_to_mongo()


## Column operator - Fix for boolean fields, e.g. WHERE study.deleted
class ColOp(_Op):

    def __init__(self, *args, **kwargs):
        super().__init__(name='COL', *args, **kwargs)
        self._identifier = SQLToken.token2sql(self.statement, self.query)

    def negate(self):
        self.is_negated = True

    def to_mongo(self):
        return {self._identifier.field: not self.is_negated}

    def evaluate(self):
        pass


class JSONOp(_Op):

    def __init__(self, *args, **kwargs):
        super().__init__(name='JSON', *args, **kwargs)
        self._identifier = SQLIdentifier(self.statement.prev_token, self.query)
        self._operator = self.statement.current_token.value
        next_tok = self.statement.next_token

        if isinstance(next_tok, Identifier):
            # Comparing against field in same doc vs a constant.
            self._is_right_constant = False
            self._right = SQLToken.token2sql(next_tok, self.query)
            if self._right.table != self._identifier.table:
                raise SQLDecodeError('Join using WHERE not supported')
        else:
            # Comparing with a constant
            self._is_right_constant = True
            set_constant(self, next_tok, unwrap_dict=False, attr='_right')

    def negate(self):
        self.is_negated = True

    @staticmethod
    def lookup_to_mongo(value):
        # For $contains queries with lookups
        if not isinstance(value, dict):
            raise SQLDecodeError(f"Invalid $contains element: {value}")
        final = defaultdict(dict)
        for k, v in value.items():
            field = k.rsplit('__', 1)

            if len(field) == 1:  # simple equality
                final[field[0]] = v
                continue

            field, lookup = '.'.join(field[:-1]), field[-1]
            '''
            Supports mongo comparison operators: https://docs.mongodb.com/manual/reference/operator/query/
            By this design, we expect the user to mix traditional ORM query with mongo operators for $contains.
            Alternatively, we could also support django-based lookups
            (https://docs.djangoproject.com/en/3.2/ref/models/querysets/#field-lookups)
            but that would require more complicated transformations here.
            '''
            if lookup in ['in', 'nin', 'lt', 'lte', 'gt', 'gte', 'eq', 'ne']:
                final[field][f'${lookup}'] = v
            else:
                raise SQLDecodeError(f'Lookup {lookup} not supported in $contains')
        return dict(final)

    def evaluate(self):
        pass

    def to_mongo(self):
        operator = self._operator
        if self._is_right_constant:
            right = parse_field(self._right, False, with_dollar=False)
        else:
            right = parse_field(self._right.field, self._right.is_array_len, with_dollar=True)
        is_negated = self.is_negated

        field = parse_field(self._identifier.field, self._identifier.is_array_len, with_dollar=False)
        field_with_dollar = parse_field(self._identifier.field, self._identifier.is_array_len, with_dollar=True)
        field_parts = self._identifier.field.rsplit('.', 1)
        if self._identifier.is_array_len and operator not in ['$exact']:
            raise SQLDecodeError(
                f"Attempting to use array length ({self._identifier}) with {operator} which doesn't make sense.")

        if operator == '$exact':
            op = '$eq' if not is_negated else '$ne'
            op2 = '$in' if not is_negated else '$nin'
            if len(field_parts) > 1 and not self._identifier.is_array_len:
                parent_field = f'${field_parts[0]}'
                return {'$expr': {
                    '$cond': [
                        {'$eq': [{'$type': parent_field}, "array"]},
                        {op2: [right, field_with_dollar]},
                        {op: [field_with_dollar, right]}
                    ]}
                }
            else:
                return {'$expr': {op: [field_with_dollar, right]}}
        elif operator == '$contains':
            if isinstance(right, dict):
                # https://docs.mongodb.com/manual/tutorial/query-array-of-documents/#a-single-nested-document-meets-multiple-query-conditions-on-nested-fields
                if is_negated:
                    return {field: {'$not': {'$elemMatch': self.lookup_to_mongo(right)}}}
                else:
                    return {field: {'$elemMatch': self.lookup_to_mongo(right)}}
            elif isinstance(right, list):
                if is_negated:
                    return {field: {'$not': {'$all': right}}}
                else:
                    return {field: {'$all': right}}
            else:
                raise SQLDecodeError(f'Invalid params for $contains: {right}')
        elif operator == '$has_key':
            return {f'{field}.{encode_key(right)}': {'$exists': not is_negated}}
        elif operator == '$has_keys':
            return {f'{field}.{encode_key(const)}': {'$exists': not is_negated} for const in right}
        elif operator == '$has_any_keys':
            return {'$or': [{f'{field}.{encode_key(const)}': {'$exists': not is_negated}} for const in right]}
        else:
            raise SQLDecodeError(f'Invalid JSONOp: {operator}')


JSON_OPERATORS = ['$exact', '$contains', '$has_key', '$has_keys', '$has_any_keys']
OPERATOR_MAP = {
    '=': '$eq',
    '>': '$gt',
    '<': '$lt',
    '>=': '$gte',
    '<=': '$lte',
    '^': '$mod',
    '*': '$multiply',
    '/': '$divide',
    '%': '$mod',
    '+': '$add',
    '-': '$subtract',
}
OPERATOR_PRECEDENCE = {
    'MATH': 11,
    'COL': 10,
    'JSON': 9,
    'IS': 8,
    'BETWEEN': 7,
    'LIKE': 6,
    'iLIKE': 6,
    'REGEXP': 6,
    'IN': 5,
    'NOT IN': 4,
    'NOT': 3,
    'AND': 2,
    'OR': 1,
    'generic': 0
}



import copy

from django.contrib import admin
from djongo.models import fields


class ModelAdmin(admin.ModelAdmin):
    DJONGO_FIELDS = (
        fields.ArrayField,
        fields.EmbeddedField,
    )

    def formfield_for_dbfield(self, db_field, request, **kwargs):
        if not isinstance(db_field, self.DJONGO_FIELDS):
            return admin.ModelAdmin.formfield_for_dbfield(
                self, db_field, request, **kwargs)

        admin_instance = ModelAdmin(db_field.model_container, admin.site)
        kwargs.setdefault('admin', admin_instance)
        kwargs.setdefault('request', request)

        for klass in db_field.__class__.mro():
            if klass in self.formfield_overrides:
                kwargs = dict(copy.deepcopy(
                    self.formfield_overrides[klass]), **kwargs)

        return db_field.formfield(**kwargs)
